(*
    Prodiams - Programmable Diagrams.
    Copyright (C) 2017 Inspirel <http://www.inspirel.com/>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

(* ::Package:: *)


BeginPackage["Prodiams`"];


diagram::usage="diagram[elements,layoutConstraints,OptionsPattern[{style\[Rule]None}]] draws a diagram including all elements and based on given layout constraints.\n"<>
"Optional style can override default settings defined in the defaultDiagramStyle association.";


pin::usage="pin[p1,p2] is a layout directive, forces two point-like structures to be in the same place.";


before::usage="before[e1,e2,OptionsPattern[{offset\[Rule]1}]] is a layout directive, puts e1 on the left of e2, with at least offset distance between their bounding boxes.";


allBefore::usage="allBefore[es,OptionsPattern[{offset\[Rule]1}]] is a layout directive, accepts a list of elements and puts every element from the list on the left of its successor, with at least offset distance between their bounding boxes.\n"<>
"allBefore[es1,es2] accepts two lists and ensures that all elements from es1 are on the left of all elements from es2.";


after::usage="after[e1,e2,OptionsPattern[{offset\[Rule]1}]] is a layout directive, puts e1 on the right of e2, with at least offset distance between their bounding boxes.";


allAfter::usage="allAfter[es,OptionsPattern[{offset\[Rule]1}]] is a layout directive, accepts a list of elements and puts every element from the list on the right of its successor, with at least offset distance between their bounding boxes.\n"<>
"allAfter[es1,es2] accepts two lists and ensures that all elements from es1 are on the right of all elements from es2.";


above::usage="above[e1,e2,OptionsPattern[{offset\[Rule]1}]] is a layout directive, puts e1 above e2, with at least offset distance between their bounding boxes.";


allAbove::usage="allAbove[es,OptionsPattern[{offset\[Rule]1}]] is a layout directive, accepts a list of elements and puts every element from the list above its successor, with at least offset distance between their bounding boxes.\n"<>
"allAbove[es1,es2] accepts two lists and ensures that all elements from es1 are above all elements from es2.";


below::usage="below[e1,e2,OptionsPattern[{offset\[Rule]1}]] is a layout directive, puts e1 below e2, with at least offset distance between their bounding boxes.";


allBelow::usage="allBelow[es,OptionsPattern[{offset\[Rule]1}]] is a layout directive, accepts a list of elements and puts every element from the list below its successor, with at least offset distance between their bounding boxes.\n"<>
"allBelow[es1,es2] accepts two lists and ensures that all elements from es1 are below all elements from es2.";


alignLeft::usage="alignLeft[e1,e2,OptionsPattern[{offset\[Rule]0}]] is a layout directive, left-aligns the bounding boxes of its arguments, with optional offset.";


allAlignLeft::usage="allAlignLeft[es] is a layout directive, accepts a list of elements and left-aligns all their bounding boxes.";


alignRight::usage="alignRight[e1,e2,OptionsPattern[{offset\[Rule]0}]] is a layout directive, right-aligns the bounding boxes of its arguments, with optional offset.";


allAlignRight::usage="allAlignRight[es] is a layout directive, accepts a list of elements and right-aligns all their bounding boxes.";


center::usage="center[e1,e2] is a layout directive, centers horizontally the bounding boxes of its arguments.";


allCenter::usage="allCenter[es] is a layout directive, accepts a list of elements and centers horizontally all their bounding boxes.";


alignTop::usage="alignTop[e1,e2,OptionsPattern[{offset\[Rule]0}]] is a layout directive, top-aligns the bounding boxes of its arguments, with optional offset.";


allAlignTop::usage="allAlignTop[es] is a layout directive, accepts a list of elements and top-aligns all their bounding boxes.";


alignBottom::usage="alignBottom[e1,e2,OptionsPattern[{offset\[Rule]0}]] is a layout directive, bottom-aligns the bounding boxes of its arguments, with optional offset.";


allAlignBottom::usage="allAlignBottom[es] is a layout directive, accepts a list of elements and bottom-aligns all their bounding boxes.";


middle::usage="middle[e1,e2] is a layout directive, centers vertically the bounding boxes of its arguments.";


allMiddle::usage="allMiddle[es] is a layout directive, accepts a list of elements and centers vertically all their bounding boxes.";


styled::usage="styled[e,OptionsPattern[{style->None}]] is a wrapper that allows to override all style settings for the given element.";


hidden::usage="hidden[e] is a wrapper that prevents the given element from being displayed (but it still takes part in layout solution).";


internalCoordinates::usage="internalCoordinates[e] returns a list of internal coordinate names for the given element.\n"<>
"This function needs to be implemented (overridden) for new element types.";


internalLayout::usage="internalLayout[e] returns a list of constraints (equalities or inequalities) involving internal coordinates of the given element.\n"<>
"This function needs to be implemented (overridden) for new element types.";


boundingBox::usage="boundingBox[e] returns a Rectangle-like coordinates of the smallest imaginary box that contains the given element. The meaning of this is not strict.\n"<>
"This function needs to be implemented (overridden) for new element types.";


specificStyle::usage="specificStyle[e] returns an association or a (possibly empty) list of rules that override default- and diagram-level style values.\n"<>
"This function needs to be implemented (overridden) for new element types.";


draw::usage="draw[e,layout,style] creates a list of graphics primitives and directives as a visual representation of the given element; layout is a list of rules with concrete values for all coordinates and style is an association with all style settings merged.";


anchor::usage="anchor[name] creates a point-like structure that has no visual representation (it is an invisible point) and is typically used to connect elements together.\n"<>
"Anchor objects are also used as intermediate values when connector points are referred.";


extent::usage="extent[p1,p2] creates an invisible box that spans two point-like structures.\n"<>
"Extents can be used as helpers for centering shapes between other, more distant objects.";


dot::usage="dot[name,OptionsPattern[{style\[Rule]None,comment->None}]] creates a point-like structure, displayed as a small disk.";


block::usage="block[name,label,OptionsPattern[{width\[Rule]5,height\[Rule]5,style\[Rule]None,comment\[Rule]None}]] creates a simple block with a label.";


line::usage="line[name,pointNames,OptionsPattern[{style\[Rule]None,comment\[Rule]None}]] creates a multi-segment line with named ends and intermediate points.";


linePoint::usage="linePoint[l,pointName] creates an anchor for the given named line point.";


inPort::usage="inPort[name,label,OptionsPattern[{style\[Rule]None,comment\[Rule]None}]] creates an input point with a label.";


outPort::usage="outPort[name,label,OptionsPattern[{style\[Rule]None,comment\[Rule]None}]] creates an output point with a label.";


connector::usage="connector[e,conName] creates an anchor for the given element's connection point.";


defaultDiagramStyle=<|
	dotStyle->Black,
	dotRadius->0.2, (* in grid units *)
	roundingRadius->0.4, (* in grid units *)
	lineStyle->Black,
	connectorLineStyle->Black,
	fillStyle->LightBlue,
	edgeStyle->Black,
	labelFunction->Function[{t,point,align},{Black,Text[Style[t,10],point,align]}],
	gridSize->15, (* size of single grid unit *)
	diagramPadding->4, (* in grid units, on all image edges *)
	printStats->False
|>;


Begin["`Private`"];


(* ::Text:: *)
(*Helper, creates a goal function in the X and Y direction for the given list of coordinates (which are supposed to be all coordinates of all elements gathered together).*)
(*The goal function is simply a sum of all coordinate values - the final solution finds a minimum of this goal function, which has the effect of contracting the diagram as close to the origin of the coordinate system, which in turn allows to minimize the size of the diagram in both X and Y directions, taking into account all layout constraints:*)


goalFunction[coordNames_]:=Total[coordNames]


(* ::Text:: *)
(*Helper, gathers all internal layout constraints from all given elements:*)


allInternalLayouts[elements_]:=Flatten[
	Map[internalLayout,elements]
]


(* ::Text:: *)
(*Helper, produces additional X and Y constraints (x>=0, y>=0) for all bounding coordinates.*)
(*These additional constraints keep the diagram in the positive quadrant of the coordinate system and together with the goal function above (sum of coordinate values) lead to small final diagrams:*)


globalConstraint[coordName_]:=coordName>=0
globalConstraints[coordNames_]:=Map[
	globalConstraint,
	coordNames
]


(* ::Text:: *)
(*Helpers, solve layout in horizontal and vertical direction.*)
(*The diagram-level and element-level constraints are joined, together with the global constraint for all coordinates.*)
(*This is the main engine's central computing function. The actual value of the goal function is ignored or the empty set is returned if no solution can be found (which can happen with contradictory layour constraints):*)


layoutSolution[elements_,coordNames_,layoutConstraints_]:=Minimize[
	{
		goalFunction[coordNames],
		Join[
			globalConstraints[coordNames],
			allInternalLayouts[elements],
			layoutConstraints
		]
	},
	coordNames,
	Integers
]

allCoordinatesLayout[elements_,coordNames_,layoutConstraints_]:=With[
	{
		solution=layoutSolution[elements,coordNames,layoutConstraints]
	},

	If[Length[solution]>0,Association[solution[[2]]],{}]
]


(* ::Text:: *)
(*Helpers, list X and Y bounding coordinates for the given bounding box.*)
(*These functions are used to help establish horizontal and vertical extents of the whole diagram:*)


horizontalElementBounds[box_]:=box[[All,1]]

verticalElementBounds[box_]:=box[[All,2]]


(* ::Text:: *)
(*Helper for computing the diagram's image size based on the extent of the whole layout.*)
(*The padding is added according to the diagram style.*)
(*This function is called separately for X and Y coordinates.*)


diagramExtent[coordValues_,diagramStyle_]:=With[
	{
		padding=diagramStyle[diagramPadding]
	},

	{Min[coordValues]-padding,Max[coordValues]+padding}
]


(* ::Text:: *)
(*Helper, gathers all internal coordinates from all given elements:*)


allInternalCoordinates[elements_]:=Flatten[
	Map[internalCoordinates,elements]
]


(* ::Text:: *)
(*Helper, normalizes diagram- or element-specific style into association:*)


specificStyleAssociation[s_]:=Association[If[s===None,{},s]]


(* ::Text:: *)
(*Main diagram function, returns the Graphics object.*)
(*The "elements" is a list of elements that take part in the diagram.*)
(*The "layoutConstraints" is a list of diagram-level constraints (typically expanded from layout directives that affect whole elements or their specific points).*)
(*The "style" is an option that overrides default diagram styles.*)
(**)
(*This function extracts all coordinate names from all given elements and joins the diagram-level constraints with element-level constraints. Then the solution is found for the global goal function and from this, the diagram (image) extent is computed, based on all element bounding boxes.*)
(*After the coordinates and found and image size is known, the actual diagram is created by calling "draw" functions for all elements in the list.*)


diagram[elements_,layoutConstraints_,OptionsPattern[{style->None}]]:=Module[
	{
		flatElements=Flatten[elements],
		boundingBoxes,
		allCoordNames,
		timedLayout,layoutTime,layout,
		timedPrimitives,graphicsTime,graphicsPrimitives,
		hBoundCoordNames,hBoundCoordValues,
		vBoundCoordNames,vBoundCoordValues,
		diagramStyle=Join[
			defaultDiagramStyle,
			specificStyleAssociation[OptionValue[style]] (* diagram-specific style *)
		],
		hExtent,vExtent,
		grid,
		hImageSize,vImageSize,
		timedImage,imageTime,image
	},

	boundingBoxes=Map[boundingBox,flatElements];
	allCoordNames=allInternalCoordinates[flatElements];

	(* Step 1. *)
	(* Find the layout solution for all elements and their constraints: *)

	timedLayout=AbsoluteTiming[
		allCoordinatesLayout[flatElements,allCoordNames,layoutConstraints]
	];
	layoutTime=Round[First[timedLayout],0.001];
	layout=Last[timedLayout];

	(* Prepare the graphics primitives based on the layout solution: *)

	timedPrimitives=AbsoluteTiming[
		Flatten[
			Table[
				draw[e,layout,
					Join[
						diagramStyle,
						specificStyleAssociation[specificStyle[e]] (* element-specific style *)
					]
				],
				{e,flatElements}
			]
		]
	];
	graphicsTime=Round[First[timedPrimitives],0.001];
	graphicsPrimitives=Last[timedPrimitives];

	(* Step 2. *)
	(* Find the diagram extent for generating proportionally-sized image: *)

	hBoundCoordNames=Flatten[
		Map[horizontalElementBounds,boundingBoxes]
	];

	vBoundCoordNames=Flatten[
		Map[verticalElementBounds,boundingBoxes]
	];

	(* Bounding boxes might involve complex expressions instead of just coordinate names, so they have to be evaluated by replacing all coordinate names with their actual values from the layout solution.
	In very simple cases (when bounding box corners are given by simple
	coordinate names), it would be sufficient to look them up in the association made from the layout, like here:

		hBoundCoordValues=Map[Association[layout],hBoundCoordNames];

	But when bounding boxes are defined in terms of more complex expressions, this lookup does not work and the evaluation can be done with term replacement, as below. *)

	(* TODO: as a performance optimization measure for large diagrams,
	think about doing this coordinate replacement by first partitioning the whole layout into smaller subsets, separate for each element (an association from elements to individual layouts), and then using such sub-layouts for replacements.
	Alternatively, check if a simple association lookup as described above works and turn to term replacement only if it fails. *)

	(* TODO: time this operation and print the absolute timing with other stats *)

	hBoundCoordValues=Replace[hBoundCoordNames,layout,All];
	vBoundCoordValues=Replace[vBoundCoordNames,layout,All];

	hExtent=diagramExtent[hBoundCoordValues,diagramStyle];
	vExtent=diagramExtent[vBoundCoordValues,diagramStyle];

	grid=diagramStyle[gridSize];
	hImageSize=(hExtent[[2]]-hExtent[[1]])*grid;
	vImageSize=(vExtent[[2]]-vExtent[[1]])*grid;

	(* Step 3. *)
	(* Create the actual diagram image. *)

	timedImage=AbsoluteTiming[
		Graphics[
			graphicsPrimitives,
			ImageSize->hImageSize,
			PlotRange->{hExtent,vExtent}
		]
	];
	imageTime=Round[First[timedImage],0.001];
	image=Last[timedImage];

	(* Step 4. *)
	(* Print statistics if requested in the diagram style options. *)

	If[diagramStyle[printStats],
		Print[Length[flatElements]," elements, ",Length[Flatten[layoutConstraints]]," constraints"];
		Print[Length[allCoordNames]," coordinates, solved in ", layoutTime, "s"];
		Print["graphics primitives generated in ",graphicsTime,"s"];
		Print["image created in ", imageTime, "s"];
	];

	image
]


(* ::Text:: *)
(*Forces two points to be in the same place:*)


pin[p1_,p2_]:=With[
	{
		x1=p1[[1]],
		y1=p1[[2]],
		x2=p2[[1]],
		y2=p2[[2]]
	},

	{
		x1==x2,
		y1==y2
	}
]


(* ::Text:: *)
(*Forces all points of e1 to be on the left of e2:*)


before[e1_,e2_,OptionsPattern[{offset->1}]]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		xMax1,xMin2,
		off=OptionValue[offset]
	},

	xMax1=box1[[2,1]];
	xMin2=box2[[1,1]];

	{
		xMax1+off<=xMin2
	}
]

allBefore[es_List,OptionsPattern[{offset->1}]]:=With[
	{
		off=OptionValue[offset]
	},

	Map[
		Function[elementPair,
			before[elementPair[[1]],elementPair[[2]],offset->off]
		],
		Partition[es,2,1]
	]
]

allBefore[es1_List,es2_List]:=Flatten[Outer[before,es1,es2]]


(* ::Text:: *)
(*Forces all points of e1 to be on the right of e2:*)


after[e1_,e2_,OptionsPattern[{offset->1}]]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		xMin1,xMax2,
		off=OptionValue[offset]
	},

	xMin1=box1[[1,1]];
	xMax2=box2[[2,1]];

	{
		xMin1>=xMax2+off
	}
]

allAfter[es_List,OptionsPattern[{offset->1}]]:=With[
	{
		off=OptionValue[offset]
	},

	Map[
		Function[elementPair,
		after[elementPair[[1]],elementPair[[2]],offset->off]
		],
		Partition[es,2,1]
	]
]

allAfter[es1_List,es2_List]:=Flatten[Outer[after,es1,es2]]


(* ::Text:: *)
(*Forces all points of e1 to be above of (higher than) e2:*)


above[e1_,e2_,OptionsPattern[{offset->1}]]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		yMin1,yMax2,
		off=OptionValue[offset]
	},

	yMin1=box1[[1,2]];
	yMax2=box2[[2,2]];

	{
		yMin1>=yMax2+off
	}
]

allAbove[es_List,OptionsPattern[{offset->1}]]:=With[
	{
		off=OptionValue[offset]
	},

	Map[
		Function[elementPair,
			above[elementPair[[1]],elementPair[[2]],offset->off]
		],
		Partition[es,2,1]
	]
]

allAbove[es1_List,es2_List]:=Flatten[Outer[above,es1,es2]]


(* ::Text:: *)
(*Forces all points of e1 to be under e2:*)


below[e1_,e2_,OptionsPattern[{offset->1}]]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		yMax1,yMin2,
		off=OptionValue[offset]
	},

	yMax1=box1[[2,2]];
	yMin2=box2[[1,2]];

	{
		yMax1+off<=yMin2
	}
]

allBelow[es_List,OptionsPattern[{offset->1}]]:=With[
	{
		off=OptionValue[offset]
	},

	Map[
		Function[elementPair,
			below[elementPair[[1]],elementPair[[2]],offset->off]
		],
		Partition[es,2,1]
	]
]

allBelow[es1_List,es2_List]:=Flatten[Outer[below,es1,es2]]


(* ::Text:: *)
(*Forces the leftmost points of e1 to be vertically aligned with leftmost points of e2:*)


alignLeft[e1_,e2_,OptionsPattern[{offset->0}]]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		xMin1,xMin2,
		off=OptionValue[offset]
	},

	xMin1=box1[[1,1]];
	xMin2=box2[[1,1]];

	{
		xMin1+off==xMin2
	}
]

allAlignLeft[es_List]:=Map[
	Function[elementPair,Apply[alignLeft,elementPair]],
	Partition[es,2,1]
]


(* ::Text:: *)
(*Forces the rightmost points of e1 to be vertically aligned with rightmost points of e2:*)


alignRight[e1_,e2_,OptionsPattern[{offset->0}]]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		xMax1,xMax2,
		off=OptionValue[offset]
	},

	xMax1=box1[[2,1]];
	xMax2=box2[[2,1]];

	{
		xMax1+off==xMax2
	}
]

allAlignRight[es_List]:=Map[
	Function[elementPair,Apply[alignRight,elementPair]],
	Partition[es,2,1]
]


(* ::Text:: *)
(*Forces the bounding boxes of e1 and e2 to be horizontally centered with regard to each other:*)


center[e1_,e2_]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		xMin1,xMax1,xMin2,xMax2
	},

	xMin1=box1[[1,1]];
	xMin2=box2[[1,1]];
	xMax1=box1[[2,1]];
	xMax2=box2[[2,1]];

	{
		xMin1+xMax1==xMin2+xMax2
	}
]

allCenter[es_List]:=Map[
	Function[elementPair,Apply[center,elementPair]],
	Partition[es,2,1]
]


(* ::Text:: *)
(*Forces the topmost points of e1 to be horizontally aligned with topmost points of e2:*)


alignTop[e1_,e2_,OptionsPattern[{offset->0}]]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		yMax1,yMax2,
		off=OptionValue[offset]
	},

	yMax1=box1[[2,2]];
	yMax2=box2[[2,2]];

	{
		yMax1+off==yMax2
	}
]

allAlignTop[es_List]:=Map[
	Function[elementPair,Apply[alignTop,elementPair]],
	Partition[es,2,1]
]


(* ::Text:: *)
(*Forces the bottom (lowest) points of e1 to be horizontally aligned with bottom points of e2:*)


alignBottom[e1_,e2_,OptionsPattern[{offset->0}]]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		yMin1,yMin2,
		off=OptionValue[offset]
	},

	yMin1=box1[[1,2]];
	yMin2=box2[[1,2]];

	{
		yMin1+off==yMin2
	}
]

allAlignBottom[es_List]:=Map[
	Function[elementPair,Apply[alignBottom,elementPair]],
	Partition[es,2,1]
]


(* ::Text:: *)
(*Forces the bounding boxes of e1 and e2 to be vertically centered with regard to each other:*)


middle[e1_,e2_]:=Module[
	{
		box1=boundingBox[e1],
		box2=boundingBox[e2],
		yMin1,yMax1,yMin2,yMax2
	},

	yMin1=box1[[1,2]];
	yMin2=box2[[1,2]];
	yMax1=box1[[2,2]];
	yMax2=box2[[2,2]];

	{
		yMin1+yMax1==yMin2+yMax2
	}
]

allMiddle[es_List]:=Map[
	Function[elementPair,Apply[middle,elementPair]],
	Partition[es,2,1]
]


(* ::Text:: *)
(*Style can apply additional style overrides for elements used in the diagram and styles set this way have the highest priority - they override all default styles, diagram-level styles and element-level styles.*)


styled[e_,OptionsPattern[{style->None}]]:=styledT[e,OptionValue[style]]

SetAttributes[styled,Listable]

internalCoordinates[s_styledT]:=internalCoordinates[s[[1]]]

internalLayout[s_styledT]:=internalLayout[s[[1]]]

boundingBox[s_styledT]:=boundingBox[s[[1]]]

specificStyle[s_styledT]:=Join[
	If[specificStyle[s[[1]]]===None,{},specificStyle[s[[1]]]],
	If[s[[2]]===None,{},s[[2]]]
]

draw[s_styledT,layout_,style_]:=draw[s[[1]],layout,style]


(* ::Text:: *)
(*Hide can hide a single element or a list of elements - the "hidden" elements are still taken into account when finding the layout solution (and therefore they "take" the same space and location on the diagram as they would normally), but are omitted when drawing.*)


hidden[e_]:=hiddenT[e]

SetAttributes[hidden,Listable]

internalCoordinates[h_hiddenT]:=internalCoordinates[h[[1]]]

internalLayout[h_hiddenT]:=internalLayout[h[[1]]]

boundingBox[h_hiddenT]:=boundingBox[h[[1]]]

specificStyle[h_hiddenT]:=specificStyle[h[[1]]]

draw[h_hiddenT,layout_,style_]:={}


anchor[name_]:=With[
	{
		anchorName=SymbolName[Unevaluated[name]]
	},

	name=Prodiams`anchorT[anchorName["x"],anchorName["y"]]
]

SetAttributes[anchor,HoldFirst]

internalCoordinates[a_anchorT]:={a[[1]],a[[2]]}

internalLayout[a_anchorT]:={}

boundingBox[a_anchorT]:={{a[[1]],a[[2]]},{a[[1]],a[[2]]}}

specificStyle[a_anchorT]:=None

draw[a_anchorT,layout_,style_]:={}


extent[p1_,p2_]:=extentT[p1,p2]

internalCoordinates[e_extentT]:=Flatten[{internalCoordinates[e[[1]]],internalCoordinates[e[[2]]]}]

internalLayout[e_extentT]:={}

boundingBox[e_extentT]:=
	{
		{Min[{e[[1]][[1]],e[[2]][[1]]}],Min[{e[[1]][[2]],e[[2]][[2]]}]},
		{Max[{e[[1]][[1]],e[[2]][[1]]}],Max[{e[[1]][[2]],e[[2]][[2]]}]}
	}

specificStyle[e_extentT]:=None

draw[e_extentT,layout_,style_]:={}


dot[name_,OptionsPattern[{style->None,comment->None}]]:=With[
	{
		dotName=SymbolName[Unevaluated[name]]
	},

	name=dotT[dotName["x"],dotName["y"],OptionValue[style],OptionValue[comment]]
]

SetAttributes[dot,HoldFirst]

internalCoordinates[d_dotT]:={d[[1]],d[[2]]}

internalLayout[d_dotT]:={}

boundingBox[d_dotT]:={{d[[1]],d[[2]]},{d[[1]],d[[2]]}}

specificStyle[d_dotT]:=d[[3]]

draw[d_dotT,layout_,style_]:=Module[
	{
		xName=d[[1]],
		yName=d[[2]],
		comment=d[[4]],
		xCenter,yCenter,
		drawing
	},

	xCenter=layout[xName];
	yCenter=layout[yName];

	drawing=Flatten[{
		style[dotStyle],Disk[{xCenter,yCenter},style[dotRadius]]
	}];

	(* comment, if specified: *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]


block[name_,label_,OptionsPattern[{width->5,height->5,style->None,comment->None}]]:=
	name=blockT[
		SymbolName[Unevaluated[name]],
		OptionValue[style],
		OptionValue[width],
		OptionValue[height],
		label,
		OptionValue[comment]
	]

SetAttributes[block,HoldFirst]

internalCoordinates[b_blockT]:=With[
	{
		name=b[[1]]
	},

	{name["xMin"],name["yMin"],name["xMax"],name["yMax"]}
]

internalLayout[b_blockT]:=With[
	{
		name=b[[1]],
		width=b[[3]],
		height=b[[4]]
	},

	{
		name["xMax"]==name["xMin"]+width,
		name["yMax"]==name["yMin"]+height
	}
]

boundingBox[b_blockT]:=With[
	{
		name=b[[1]]
	},

	{{name["xMin"],name["yMin"]},{name["xMax"],name["yMax"]}}
]

specificStyle[b_blockT]:=b[[2]]

draw[b_blockT,layout_,style_]:=Module[
	{
		name=b[[1]],
		label=b[[5]],
		comment=b[[6]],
		xMin,yMin,
		xMax,yMax,
		xCenter,yCenter,
		drawing
	},

	xMin=layout[name["xMin"]];
	yMin=layout[name["yMin"]];
	xMax=layout[name["xMax"]];
	yMax=layout[name["yMax"]];

	xCenter=Mean[{xMin,xMax}];
	yCenter=Mean[{yMin,yMax}];

	drawing=Flatten[{
		style[fillStyle],EdgeForm[style[edgeStyle]],
		Rectangle[{xMin,yMin},{xMax,yMax},RoundingRadius->style[roundingRadius]],
		style[labelFunction][label,{xCenter,yCenter},{0,0}]
	}];

	(* comment, if specified: *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]


line[name_,pointNames_List,OptionsPattern[{style->None,comment->None}]]:=
	name=lineT[
		SymbolName[Unevaluated[name]],
		pointNames,
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[line,HoldFirst]

makeInternalCoordinates[l_lineT,suffix_String]:=With[
	{
		name=l[[1]],
		pointNames=l[[2]]
	},

	Map[
		Function[pn,name[pn][suffix]],
		pointNames
	]
]

horizontalInternalCoordinates[l_lineT]:=makeInternalCoordinates[l,"x"]
verticalInternalCoordinates[l_lineT]:=makeInternalCoordinates[l,"y"]

internalCoordinates[l_lineT]:=Riffle[
	horizontalInternalCoordinates[l],
	verticalInternalCoordinates[l]
]

internalLayout[l_lineT]:={}

boundingBox[l_lineT]:=With[
	{
		hCoordNames=horizontalInternalCoordinates[l],
		vCoordNames=verticalInternalCoordinates[l]
	},

	{
		{Min[hCoordNames],Min[vCoordNames]},
		{Max[hCoordNames],Max[vCoordNames]}
	}
]

specificStyle[l_lineT]:=l[[3]]

draw[l_lineT,layout_,style_]:=Module[
	{
		name=l[[1]],
		comment=l[[4]],
		coordValues,
		drawing
	},

	coordValues=Map[
		Function[coordNamePair,
			{layout[coordNamePair[[1]]],layout[coordNamePair[[2]]]}
		],

		(* partition a flat list of coordinate names into a list of {x,y} pairs *)
		Partition[internalCoordinates[l],2]
	];

	drawing={style[lineStyle],Line[coordValues]};

	(* comment, if specified: *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]

linePoint[l_lineT,pointName_]:=With[
	{
		name=l[[1]]
	},

	anchorT[name[pointName]["x"],name[pointName]["y"]]
]

SetAttributes[linePoint,Listable]


(* ::Text:: *)
(*Fixed common value, radius of the circle used in input and out port shapes:*)


portCircleRadius=0.4; (* in grid units *)


inPort[name_,label_,OptionsPattern[{style->None,comment->None}]]:=With[
	{
		inPortName=SymbolName[Unevaluated[name]]
	},

	name=inPortT[
		inPortName["x"],
		inPortName["y"],
		label,
		OptionValue[style],
		OptionValue[comment]
	]
]

SetAttributes[inPort,HoldFirst]

internalCoordinates[p_inPortT]:={p[[1]],p[[2]]}

internalLayout[p_inPortT]:={}

boundingBox[p_inPortT]:={{p[[1]]-2,p[[2]]},{p[[1]],p[[2]]}}

specificStyle[p_inPortT]:=p[[4]]

draw[p_inPortT,layout_,style_]:=Module[
	{
		xName=p[[1]],
		yName=p[[2]],
		label=p[[3]],
		comment=p[[4]],
		x,y,
		drawing
	},

	x=layout[xName];
	y=layout[yName];

	drawing=Flatten[{
		style[lineStyle],
		Circle[{x-2,y},portCircleRadius],
		style[connectorLineStyle],
		Line[{{x-2+portCircleRadius,y},{x,y}}],
		style[labelFunction][label,{x-2-2*portCircleRadius,y},{1,0}]
	}];

	(* comment, if specified: *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]


outPort[name_,label_,OptionsPattern[{style->None,comment->None}]]:=With[
	{
		outPortName=SymbolName[Unevaluated[name]]
	},

	name=outPortT[
		outPortName["x"],
		outPortName["y"],
		label,
		OptionValue[style],
		OptionValue[comment]
	]
]

SetAttributes[outPort,HoldFirst]

internalCoordinates[p_outPortT]:={p[[1]],p[[2]]}

internalLayout[p_outPortT]:={}

boundingBox[p_outPortT]:={{p[[1]],p[[2]]},{p[[1]]+2,p[[2]]}}

specificStyle[p_outPortT]:=p[[4]]

draw[p_outPortT,layout_,style_]:=Module[
	{
		xName=p[[1]],
		yName=p[[2]],
		label=p[[3]],
		comment=p[[4]],
		x,y,
		drawing
	},

	x=layout[xName];
	y=layout[yName];

	drawing=Flatten[{
		style[lineStyle],
		Circle[{x+2,y},portCircleRadius],
		style[connectorLineStyle],
		Line[{{x+2-portCircleRadius,y},{x,y}}],
		style[labelFunction][label,{x+2+2*portCircleRadius,y},{-1,0}]
	}];

	(* comment, if specified: *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]


End[];


EndPackage[];
